#
# Copyright (c) 2006, 2007 Canonical
#
# Written by Gustavo Niemeyer <gustavo@niemeyer.net>
#
# This file is part of Storm Object Relational Mapper.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Basic database interfacing mechanisms for Storm.

This is the common code for database support; specific databases are
supported in modules in L{storm.databases}.
"""

from collections.abc import Callable
from functools import wraps

from storm.expr import Expr, State, compile
# Circular import: imported at the end of the module.
# from storm.tracer import trace
from storm.variables import Variable
from storm.xid import Xid
from storm.exceptions import (
    ClosedError, ConnectionBlockedError, DatabaseError, DisconnectionError,
    Error, ProgrammingError, wrap_exceptions)
from storm.uri import URI
import storm


__all__ = ["Database", "Connection", "Result",
           "convert_param_marks", "create_database", "register_scheme"]


STATE_CONNECTED = 1
STATE_DISCONNECTED = 2
STATE_RECONNECT = 3


class Result:
    """A representation of the results from a single SQL statement."""

    _closed = False

    def __init__(self, connection, raw_cursor):
        self._connection = connection # Ensures deallocation order.
        self._raw_cursor = raw_cursor
        if raw_cursor.arraysize == 1:
            # Default of 1 is silly.
            self._raw_cursor.arraysize = 10

    def __del__(self):
        """Close the cursor."""
        try:
            self.close()
        except:
            pass

    def close(self):
        """Close the underlying raw cursor, if it hasn't already been closed.
        """
        if not self._closed:
            self._closed = True
            self._raw_cursor.close()
            self._raw_cursor = None

    def get_one(self):
        """Fetch one result from the cursor.

        The result will be converted to an appropriate format via
        L{from_database}.

        @raise DisconnectionError: Raised when the connection is lost.
            Reconnection happens automatically on rollback.

        @return: A converted row or None, if no data is left.
        """
        row = self._connection._check_disconnect(self._raw_cursor.fetchone)
        if row is not None:
            return tuple(self.from_database(row))
        return None

    def get_all(self):
        """Fetch all results from the cursor.

        The results will be converted to an appropriate format via
        L{from_database}.

        @raise DisconnectionError: Raised when the connection is lost.
            Reconnection happens automatically on rollback.
        """
        result = self._connection._check_disconnect(self._raw_cursor.fetchall)
        if result:
            return [tuple(self.from_database(row)) for row in result]
        return result

    def __iter__(self):
        """Yield all results, one at a time.

        The results will be converted to an appropriate format via
        L{from_database}.

        @raise DisconnectionError: Raised when the connection is lost.
            Reconnection happens automatically on rollback.
        """
        while True:
            results = self._connection._check_disconnect(
                self._raw_cursor.fetchmany)
            if not results:
                break
            for result in results:
                yield tuple(self.from_database(result))

    @property
    def rowcount(self):
        """
        See PEP 249 for further details on rowcount.

        @return: the number of affected rows, or None if the database
            backend does not provide this information. Return value
            is undefined if all results have not yet been retrieved.
        """
        if self._raw_cursor.rowcount == -1:
            return None
        return self._raw_cursor.rowcount

    def get_insert_identity(self, primary_columns, primary_variables):
        """Get a query which will return the row that was just inserted.

        This must be overridden in database-specific subclasses.

        @rtype: L{storm.expr.Expr}
        """
        raise NotImplementedError

    @staticmethod
    def set_variable(variable, value):
        """Set the given variable's value from the database."""
        variable.set(value, from_db=True)

    @staticmethod
    def from_database(row):
        """Convert a row fetched from the database to an agnostic format.

        This method is intended to be overridden in subclasses, but
        not called externally.

        If there are any peculiarities in the datatypes returned from
        a database backend, this method should be overridden in the
        backend subclass to convert them.
        """
        return row


class CursorWrapper:
    """A DB-API cursor, wrapping exceptions as StormError instances."""

    def __init__(self, cursor, database):
        super().__setattr__('_cursor', cursor)
        super().__setattr__('_database', database)

    def __getattr__(self, name):
        attr = getattr(self._cursor, name)
        if isinstance(attr, Callable):
            @wraps(attr)
            def wrapper(*args, **kwargs):
                with wrap_exceptions(self._database):
                    return attr(*args, **kwargs)

            return wrapper
        else:
            return attr

    def __setattr__(self, name, value):
        return setattr(self._cursor, name, value)

    def __iter__(self):
        with wrap_exceptions(self._database):
            yield from self._cursor

    def __enter__(self):
        return self

    def __exit__(self, type_, value, tb):
        with wrap_exceptions(self._database):
            self.close()


class ConnectionWrapper:
    """A DB-API connection, wrapping exceptions as StormError instances."""

    def __init__(self, connection, database):
        self.__dict__['_connection'] = connection
        self.__dict__['_database'] = database

    def __getattr__(self, name):
        attr = getattr(self._connection, name)
        if isinstance(attr, Callable):
            @wraps(attr)
            def wrapper(*args, **kwargs):
                with wrap_exceptions(self._database):
                    return attr(*args, **kwargs)

            return wrapper
        else:
            return attr

    def __setattr__(self, name, value):
        return setattr(self._connection, name, value)

    def __enter__(self):
        return self

    def __exit__(self, type_, value, tb):
        with wrap_exceptions(self._database):
            if type_ is None and value is None and tb is None:
                self.commit()
            else:
                self.rollback()

    def cursor(self):
        with wrap_exceptions(self._database):
            return CursorWrapper(self._connection.cursor(), self._database)


class Connection:
    """A connection to a database.

    @cvar result_factory: A callable which takes this L{Connection}
        and the backend cursor and returns an instance of L{Result}.
    @type param_mark: C{str}
    @cvar param_mark: The dbapi paramstyle that the database backend expects.
    @type compile: L{storm.expr.Compile}
    @cvar compile: The compiler to use for connections of this type.
    """

    result_factory = Result
    param_mark = "?"
    compile = compile

    _blocked = False
    _closed = False
    _two_phase_transaction = False  # If True, a two-phase transaction has
                                    # been started with begin()
    _state = STATE_CONNECTED

    def __init__(self, database, event=None):
        self._database = database # Ensures deallocation order.
        self._event = event
        self._raw_connection = self._database.raw_connect()

    def __del__(self):
        """Close the connection."""
        try:
            self.close()
        except:
            pass

    def block_access(self):
        """Block access to the connection.

        Attempts to execute statements or commit a transaction will
        result in a C{ConnectionBlockedError} exception.  Rollbacks
        are permitted as that operation is often used in case of
        failures.
        """
        self._blocked = True

    def unblock_access(self):
        """Unblock access to the connection."""
        self._blocked = False

    def execute(self, statement, params=None, noresult=False):
        """Execute a statement with the given parameters.

        @type statement: L{Expr} or C{str}
        @param statement: The statement to execute. It will be
            compiled if necessary.
        @param noresult: If True, no result will be returned.

        @raise ConnectionBlockedError: Raised if access to the connection
            has been blocked with L{block_access}.
        @raise DisconnectionError: Raised when the connection is lost.
            Reconnection happens automatically on rollback.

        @return: The result of C{self.result_factory}, or None if
            C{noresult} is True.
        """
        if self._closed:
            raise ClosedError("Connection is closed")
        if self._blocked:
            raise ConnectionBlockedError("Access to connection is blocked")
        if self._event:
            self._event.emit("register-transaction")
        self._ensure_connected()
        if isinstance(statement, Expr):
            if params is not None:
                raise ValueError("Can't pass parameters with expressions")
            state = State()
            statement = self.compile(statement, state)
            params = state.parameters
        statement = convert_param_marks(statement, "?", self.param_mark)
        raw_cursor = self.raw_execute(statement, params)
        if noresult:
            self._check_disconnect(raw_cursor.close)
            return None
        return self.result_factory(self, raw_cursor)

    def close(self):
        """Close the connection if it is not already closed."""
        if not self._closed:
            self._closed = True
            if self._raw_connection is not None:
                self._raw_connection.close()
                self._raw_connection = None

    def begin(self, xid):
        """Begin a two-phase transaction."""
        if self._two_phase_transaction:
            raise ProgrammingError("begin cannot be used inside a transaction")
        self._ensure_connected()
        raw_xid = self._raw_xid(xid)
        self._check_disconnect(self._raw_connection.tpc_begin, raw_xid)
        self._two_phase_transaction = True

    def prepare(self):
        """Run the prepare phase of a two-phase transaction."""
        if not self._two_phase_transaction:
            raise ProgrammingError("prepare must be called inside a two-phase "
                                   "transaction")
        self._check_disconnect(self._raw_connection.tpc_prepare)

    def commit(self, xid=None):
        """Commit the connection.

        @param xid: Optionally the L{Xid} of a previously prepared
             transaction to commit. This form should be called outside
             of a transaction, and is intended for use in recovery.

        @raise ConnectionBlockedError: Raised if access to the connection
            has been blocked with L{block_access}.
        @raise DisconnectionError: Raised when the connection is lost.
            Reconnection happens automatically on rollback.

        """
        try:
            self._ensure_connected()
            if xid:
                raw_xid = self._raw_xid(xid)
                self._check_disconnect(self._raw_connection.tpc_commit, raw_xid)
            elif self._two_phase_transaction:
                self._check_disconnect(self._raw_connection.tpc_commit)
                self._two_phase_transaction = False
            else:
                self._check_disconnect(self._raw_connection.commit)
        finally:
            self._check_disconnect(trace, "connection_commit", self, xid)

    def recover(self):
        """Return a list of L{Xid}\\ s representing pending transactions."""
        self._ensure_connected()
        raw_xids = self._check_disconnect(self._raw_connection.tpc_recover)
        return [Xid(raw_xid[0], raw_xid[1], raw_xid[2])
                for raw_xid in raw_xids]

    def rollback(self, xid=None):
        """Rollback the connection.

        @param xid: Optionally the L{Xid} of a previously prepared
             transaction to rollback. This form should be called outside
             of a transaction, and is intended for use in recovery.
        """
        try:
            if self._state == STATE_CONNECTED:
                try:
                    if xid:
                        raw_xid = self._raw_xid(xid)
                        self._raw_connection.tpc_rollback(raw_xid)
                    elif self._two_phase_transaction:
                        self._raw_connection.tpc_rollback()
                    else:
                        self._raw_connection.rollback()
                except Error as exc:
                    if self.is_disconnection_error(exc):
                        self._raw_connection = None
                        self._state = STATE_RECONNECT
                        self._two_phase_transaction = False
                    else:
                        raise
                else:
                    self._two_phase_transaction = False
            else:
                self._two_phase_transaction = False
                self._state = STATE_RECONNECT
        finally:
            self._check_disconnect(trace, "connection_rollback", self, xid)

    @staticmethod
    def to_database(params):
        """Convert some parameters into values acceptable to a database backend.

        It is acceptable to override this method in subclasses, but it
        is not intended to be used externally.

        This delegates conversion to any
        L{Variable <storm.variable.Variable>}\\ s in the parameter list, and
        passes through all other values untouched.
        """
        for param in params:
            if isinstance(param, Variable):
                yield param.get(to_db=True)
            else:
                yield param

    def build_raw_cursor(self):
        """Get a new dbapi cursor object.

        It is acceptable to override this method in subclasses, but it
        is not intended to be called externally.
        """
        return self._raw_connection.cursor()

    def raw_execute(self, statement, params=None):
        """Execute a raw statement with the given parameters.

        It's acceptable to override this method in subclasses, but it
        is not intended to be called externally.

        If the global C{DEBUG} is True, the statement will be printed
        to standard out.

        @return: The dbapi cursor object, as fetched from L{build_raw_cursor}.
        """
        raw_cursor = self._check_disconnect(self.build_raw_cursor)
        self._prepare_execution(raw_cursor, params, statement)
        args = self._execution_args(params, statement)
        self._run_execution(raw_cursor, args, params, statement)
        return raw_cursor

    def _execution_args(self, params, statement):
        """Get the appropriate statement execution arguments."""
        if params:
            args = (statement, tuple(self.to_database(params)))
        else:
            args = (statement,)
        return args

    def _run_execution(self, raw_cursor, args, params, statement):
        """Complete the statement execution, along with result reports."""
        try:
            self._check_disconnect(raw_cursor.execute, *args)
        except Exception as error:
            self._check_disconnect(
                trace, "connection_raw_execute_error", self, raw_cursor,
                statement, params or (), error)
            raise
        else:
            self._check_disconnect(
                trace, "connection_raw_execute_success", self, raw_cursor,
                statement, params or ())

    def _prepare_execution(self, raw_cursor, params, statement):
        """Prepare the statement execution to be run."""
        try:
            self._check_disconnect(
                trace, "connection_raw_execute", self, raw_cursor,
                statement, params or ())
        except Exception as error:
            self._check_disconnect(
                trace, "connection_raw_execute_error", self, raw_cursor,
                statement, params or (), error)
            raise

    def _ensure_connected(self):
        """Ensure that we are connected to the database.

        If the connection is marked as dead, or if we can't reconnect,
        then raise DisconnectionError.
        """
        if self._blocked:
            raise ConnectionBlockedError("Access to connection is blocked")
        if self._state == STATE_CONNECTED:
            return
        elif self._state == STATE_DISCONNECTED:
            raise DisconnectionError("Already disconnected")
        elif self._state == STATE_RECONNECT:
            try:
                self._raw_connection = self._database.raw_connect()
            except DatabaseError as exc:
                self._state = STATE_DISCONNECTED
                self._raw_connection = None
                raise DisconnectionError(str(exc))
            else:
                self._state = STATE_CONNECTED

    def is_disconnection_error(self, exc, extra_disconnection_errors=()):
        """Check whether an exception represents a database disconnection.

        This should be overridden by backends to detect whichever
        exception values are used to represent this condition.
        """
        return False

    def _raw_xid(self, xid):
        """Return a raw xid from the given high-level L{Xid} object."""
        return self._raw_connection.xid(xid.format_id,
                                        xid.global_transaction_id,
                                        xid.branch_qualifier)

    def _check_disconnect(self, function, *args, **kwargs):
        """Run the given function, checking for database disconnections."""
        # Allow the caller to specify additional exception types that
        # should be treated as possible disconnection errors.
        extra_disconnection_errors = kwargs.pop(
            'extra_disconnection_errors', ())
        try:
            return function(*args, **kwargs)
        except Exception as exc:
            if self.is_disconnection_error(exc, extra_disconnection_errors):
                self._state = STATE_DISCONNECTED
                self._raw_connection = None
                raise DisconnectionError(str(exc))
            else:
                raise

    def preset_primary_key(self, primary_columns, primary_variables):
        """Process primary variables before an insert happens.

        This method may be overwritten by backends to implement custom
        changes in primary variables before an insert happens.
        """


class Database:
    """A database that can be connected to.

    This should be subclassed for individual database backends.

    @cvar connection_factory: A callable which will take this database
        and should return an instance of L{Connection}.
    """

    connection_factory = Connection

    def __init__(self, uri=None):
        self._uri = uri
        self._exception_types = {}

    def get_uri(self):
        """Return the URI object this database was created with."""
        return self._uri

    def connect(self, event=None):
        """Create a connection to the database.

        It calls C{self.connection_factory} to allow for ease of
        customization.

        @param event: The event system to broadcast messages with. If
            not specified, then no events will be broadcast.

        @return: An instance of L{Connection}.
        """
        return self.connection_factory(self, event)

    def raw_connect(self):
        """Create a raw database connection.

        This is used by L{Connection} objects to connect to the
        database.  It should be overriden in subclasses to do any
        database-specific connection setup.

        @return: A DB-API connection object.
        """
        raise NotImplementedError

    @property
    def _exception_module(self):
        """The module where appropriate DB-API exception types are defined.

        Subclasses should set this if they support re-raising DB-API
        exceptions as StormError instances.
        """
        return None

    def _make_combined_exception_type(self, wrapper_type, dbapi_type):
        """Make a combined exception based on both DB-API and Storm.

        Storm historically defined its own exception types as ABCs and
        registered the DB-API exception types as virtual subclasses.
        However, this doesn't work properly in Python 3
        (https://bugs.python.org/issue12029).  Instead, we create and cache
        subclass-specific exception types that inherit from both StormError
        and the DB-API exception type, allowing code that catches either
        StormError (or subclasses) or the specific DB-API exceptions to keep
        working.

        @type wrapper_type: L{type}
        @param wrapper_type: The type of the wrapper exception to create; a
            subclass of L{StormError}.
        @type dbapi_type: L{type}
        @param dbapi_type: The type of the DB-API exception.

        @return: The combined exception type.
        """
        if dbapi_type.__name__ not in self._exception_types:
            self._exception_types[dbapi_type.__name__] = type(
                dbapi_type.__name__, (dbapi_type, wrapper_type), {})
        return self._exception_types[dbapi_type.__name__]

    def _wrap_exception(self, wrapper_type, exception):
        """Wrap a DB-API exception as a StormError instance.

        This constructs a wrapper exception with the same C{args} as the
        DB-API exception.  Subclasses may override this to set additional
        attributes on the wrapper exception.

        @type wrapper_type: L{type}
        @param wrapper_type: The type of the wrapper exception to create; a
            subclass of L{StormError}.
        @type exception: L{Exception}
        @param exception: The DB-API exception to wrap.

        @return: The wrapped exception; an instance of L{StormError}.
        """
        return self._make_combined_exception_type(
            wrapper_type, exception.__class__)(*exception.args)


def convert_param_marks(statement, from_param_mark, to_param_mark):
    # TODO: Add support for $foo$bar$foo$ literals.
    if from_param_mark == to_param_mark or from_param_mark not in statement:
        return statement
    tokens = statement.split("'")
    for i in range(0, len(tokens), 2):
        tokens[i] = tokens[i].replace(from_param_mark, to_param_mark)
    return "'".join(tokens)


_database_schemes = {}

def register_scheme(scheme, factory):
    """Register a handler for a new database URI scheme.

    @param scheme: the database URI scheme
    @param factory: a function taking a URI instance and returning a database.
    """
    _database_schemes[scheme] = factory


def create_database(uri):
    """Create a database instance.

    @param uri: An URI instance, or a string describing the URI. Some examples:

    "sqlite:"
        An in memory sqlite database.

    "sqlite:example.db"
        A SQLite database called example.db

    "postgres:test"
        The database 'test' from the local postgres server.

    "postgres://user:password@host/test"
        The database test on machine host with supplied user credentials,
        using postgres.

    "anything:..."
        Where 'anything' has previously been registered with
        L{register_scheme}.
    """
    if isinstance(uri, str):
        uri = URI(uri)
    if uri.scheme in _database_schemes:
        factory = _database_schemes[uri.scheme]
    else:
        module = __import__("%s.databases.%s" % (storm.__name__, uri.scheme),
                            None, None, [""])
        factory = module.create_from_uri
    return factory(uri)

# Deal with circular import.
from storm.tracer import trace



#
# Copyright (c) 2006, 2007 Canonical
#
# Written by Gustavo Niemeyer <gustavo@niemeyer.net>
#
# This file is part of Storm Object Relational Mapper.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import errno
import os
import select
import socket
import socketserver
import threading


TIMEOUT = 0.1


class ProxyRequestHandler(socketserver.BaseRequestHandler):
    """A request handler that proxies traffic to another TCP port."""

    def __init__(self, request, client_address, server):
        self._generation = server._generation
        socketserver.BaseRequestHandler.__init__(
            self, request, client_address, server)

    def handle(self):
        if isinstance(self.server.proxy_dest, (bytes, str)):
            dst = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        else:
            dst = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        dst.connect(self.server.proxy_dest)

        readers = [self.request, dst]
        while readers:
            rlist, wlist, xlist = select.select(readers, [], [], TIMEOUT)

            # If the server generation has been incremented, close the
            # connection.
            if self._generation != self.server._generation:
                return

            if self.request in rlist:
                chunk = os.read(self.request.fileno(), 1024)
                try:
                    dst.send(chunk)
                except OSError as e:
                    if e.errno == errno.EPIPE:
                        return
                    raise
                if chunk == "":
                    readers.remove(self.request)
                    dst.shutdown(socket.SHUT_WR)

            if dst in rlist:
                try:
                    chunk = os.read(dst.fileno(), 1024)
                except OSError as e:
                    if e.errno == errno.ECONNRESET:
                        return
                    raise
                self.request.send(chunk)
                if chunk == "":
                    readers.remove(dst)
                    self.request.shutdown(socket.SHUT_WR)


class ProxyTCPServer(socketserver.ThreadingTCPServer):

    allow_reuse_address = True

    def __init__(self, proxy_dest):
        socketserver.ThreadingTCPServer.__init__(
            self, ("127.0.0.1", 0), ProxyRequestHandler)
        # Python 2.4 doesn't retrieve the socket details, so record
        # them here.  We need to do this so we can recreate the socket
        # with the same address later.
        self.server_address = self.socket.getsockname()

        self.proxy_dest = proxy_dest
        self._start_lock = threading.Lock()
        self._thread = None
        self._generation = 0
        self._running = False
        self.start()

    def __del__(self):
        self.close()

    def close(self):
        if self._running:
            self.stop()

    def start(self):
        assert not self._running, "Server should not be running"
        self._thread = threading.Thread(target=self._run)
        self._thread.setDaemon(True)

        self._running = True
        self._start_lock.acquire()
        self._thread.start()
        # Wait for server to start
        self._start_lock.acquire()
        self._start_lock.release()

    def _run(self):
        self.server_activate()
        self.socket.settimeout(TIMEOUT)
        self._start_lock.release()
        while self._running:
            try:
                self.handle_request()
            except socket.timeout:
                pass

    def stop(self):
        assert self._running, "Server should be running"
        # Increment server generation, and wait for thread to stop.
        self._generation += 1
        self._running = False
        self._thread.join()

        # Recreate socket, to kill listen queue.  As we've allowed
        # address reuse, this should work.
        self.socket.close()
        self.socket = socket.socket(self.address_family,
                                    self.socket_type)
        self.server_bind()

    def restart(self):
        self.stop()
        self.start()



#
# Copyright (c) 2006, 2007 Canonical
#
# Written by Gustavo Niemeyer <gustavo@niemeyer.net>
#
# This file is part of Storm Object Relational Mapper.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from weakref import ref
import gc

from storm.exceptions import ClassInfoError
from storm.properties import Property
from storm.variables import Variable
from storm.expr import Undef, Select, compile
from storm.info import *
from storm.tests.helper import TestHelper


class Wrapper:

    def __init__(self, obj):
        self.obj = obj

    __storm_object_info__ = property(lambda self:
                                     self.obj.__storm_object_info__)


class GetTest(TestHelper):

    def setUp(self):
        TestHelper.setUp(self)
        class Class:
            __storm_table__ = "table"
            prop1 = Property("column1", primary=True)
        self.Class = Class
        self.obj = Class()

    def test_get_cls_info(self):
        cls_info = get_cls_info(self.Class)
        self.assertTrue(isinstance(cls_info, ClassInfo))
        self.assertTrue(cls_info is get_cls_info(self.Class))

    def test_get_obj_info(self):
        obj_info = get_obj_info(self.obj)
        self.assertTrue(isinstance(obj_info, ObjectInfo))
        self.assertTrue(obj_info is get_obj_info(self.obj))

    def test_get_obj_info_on_obj_info(self):
        obj_info = get_obj_info(self.obj)
        self.assertTrue(get_obj_info(obj_info) is obj_info)

    def test_set_obj_info(self):
        obj_info1 = get_obj_info(self.obj)
        obj_info2 = ObjectInfo(self.obj)
        self.assertEqual(get_obj_info(self.obj), obj_info1)
        set_obj_info(self.obj, obj_info2)
        self.assertEqual(get_obj_info(self.obj), obj_info2)


class ClassInfoTest(TestHelper):

    def setUp(self):
        TestHelper.setUp(self)
        class Class:
            __storm_table__ = "table"
            prop1 = Property("column1", primary=True)
            prop2 = Property("column2")
        self.Class = Class
        self.cls_info = get_cls_info(Class)

    def test_invalid_class(self):
        class Class: pass
        self.assertRaises(ClassInfoError, ClassInfo, Class)

    def test_cls(self):
        self.assertEqual(self.cls_info.cls, self.Class)

    def test_columns(self):
        self.assertEqual(self.cls_info.columns,
                         (self.Class.prop1, self.Class.prop2))

    def test_table(self):
        self.assertEqual(self.cls_info.table.name, "table")

    def test_primary_key(self):
        # Can't use == for props.
        self.assertTrue(self.cls_info.primary_key[0] is self.Class.prop1)
        self.assertEqual(len(self.cls_info.primary_key), 1)

    def test_primary_key_with_attribute(self):
        class SubClass(self.Class):
            __storm_primary__ = "prop2"

        cls_info = get_cls_info(SubClass)

        self.assertTrue(cls_info.primary_key[0] is SubClass.prop2)
        self.assertEqual(len(self.cls_info.primary_key), 1)

    def test_primary_key_composed(self):
        class Class:
            __storm_table__ = "table"
            prop1 = Property("column1", primary=2)
            prop2 = Property("column2", primary=1)
        cls_info = ClassInfo(Class)

        # Can't use == for props, since they're columns.
        self.assertTrue(cls_info.primary_key[0] is Class.prop2)
        self.assertTrue(cls_info.primary_key[1] is Class.prop1)
        self.assertEqual(len(cls_info.primary_key), 2)

    def test_primary_key_composed_with_attribute(self):
        class Class:
            __storm_table__ = "table"
            __storm_primary__ = "prop2", "prop1"
            # Define primary=True to ensure that the attribute
            # has precedence.
            prop1 = Property("column1", primary=True)
            prop2 = Property("column2")
        cls_info = ClassInfo(Class)

        # Can't use == for props, since they're columns.
        self.assertTrue(cls_info.primary_key[0] is Class.prop2)
        self.assertTrue(cls_info.primary_key[1] is Class.prop1)
        self.assertEqual(len(cls_info.primary_key), 2)

    def test_primary_key_composed_duplicated(self):
        class Class:
            __storm_table__ = "table"
            prop1 = Property("column1", primary=True)
            prop2 = Property("column2", primary=True)
        self.assertRaises(ClassInfoError, ClassInfo, Class)

    def test_primary_key_missing(self):
        class Class:
            __storm_table__ = "table"
            prop1 = Property("column1")
            prop2 = Property("column2")
        self.assertRaises(ClassInfoError, ClassInfo, Class)

    def test_primary_key_attribute_missing(self):
        class Class:
            __storm_table__ = "table"
            __storm_primary__ = ()
            prop1 = Property("column1", primary=True)
            prop2 = Property("column2")
        self.assertRaises(ClassInfoError, ClassInfo, Class)

    def test_primary_key_pos(self):
        class Class:
            __storm_table__ = "table"
            prop1 = Property("column1", primary=2)
            prop2 = Property("column2")
            prop3 = Property("column3", primary=1)
        cls_info = ClassInfo(Class)
        self.assertEqual(cls_info.primary_key_pos, (2, 0))


class ObjectInfoTest(TestHelper):

    def setUp(self):
        TestHelper.setUp(self)
        class Class:
            __storm_table__ = "table"
            prop1 = Property("column1", primary=True)
            prop2 = Property("column2")
        self.Class = Class
        self.obj = Class()
        self.obj_info = get_obj_info(self.obj)
        self.cls_info = get_cls_info(Class)
        self.variable1 = self.obj_info.variables[Class.prop1]
        self.variable2 = self.obj_info.variables[Class.prop2]

    def test_hashing(self):
        self.assertEqual(hash(self.obj_info), hash(self.obj_info))

    def test_equals(self):
        obj_info1 = self.obj_info
        obj_info2 = get_obj_info(self.Class())
        self.assertFalse(obj_info1 == obj_info2)

    def test_not_equals(self):
        obj_info1 = self.obj_info
        obj_info2 = get_obj_info(self.Class())
        self.assertTrue(obj_info1 != obj_info2)

    def test_dict_subclass(self):
        self.assertTrue(isinstance(self.obj_info, dict))

    def test_variables(self):
        self.assertTrue(isinstance(self.obj_info.variables, dict))

        for column in self.cls_info.columns:
            variable = self.obj_info.variables.get(column)
            self.assertTrue(isinstance(variable, Variable))
            self.assertTrue(variable.column is column)

        self.assertEqual(len(self.obj_info.variables),
                         len(self.cls_info.columns))

    def test_variable_has_validator_object_factory(self):
        args = []
        def validator(obj, attr, value):
            args.append((obj, attr, value))
        class Class:
            __storm_table__ = "table"
            prop = Property(primary=True,
                            variable_kwargs={"validator": validator})

        obj = Class()
        get_obj_info(obj).variables[Class.prop].set(123)

        self.assertEqual(args, [(obj, "prop", 123)])

    def test_primary_vars(self):
        self.assertTrue(isinstance(self.obj_info.primary_vars, tuple))

        for column, variable in zip(self.cls_info.primary_key,
                                    self.obj_info.primary_vars):
            self.assertEqual(self.obj_info.variables.get(column),
                             variable)

        self.assertEqual(len(self.obj_info.primary_vars),
                         len(self.cls_info.primary_key))

    def test_checkpoint(self):
        self.obj.prop1 = 10
        self.obj_info.checkpoint()
        self.assertEqual(self.obj.prop1, 10)
        self.assertEqual(self.variable1.has_changed(), False)
        self.obj.prop1 = 20
        self.assertEqual(self.obj.prop1, 20)
        self.assertEqual(self.variable1.has_changed(), True)
        self.obj_info.checkpoint()
        self.assertEqual(self.obj.prop1, 20)
        self.assertEqual(self.variable1.has_changed(), False)
        self.obj.prop1 = 20
        self.assertEqual(self.obj.prop1, 20)
        self.assertEqual(self.variable1.has_changed(), False)

    def test_add_change_notification(self):
        changes1 = []
        changes2 = []
        def object_changed1(obj_info, variable, old_value, new_value, fromdb):
            changes1.append((1, obj_info, variable,
                             old_value, new_value, fromdb))
        def object_changed2(obj_info, variable, old_value, new_value, fromdb):
            changes2.append((2, obj_info, variable,
                             old_value, new_value, fromdb))

        self.obj_info.checkpoint()
        self.obj_info.event.hook("changed", object_changed1)
        self.obj_info.event.hook("changed", object_changed2)

        self.obj.prop2 = 10
        self.obj.prop1 = 20

        self.assertEqual(changes1,
                      [(1, self.obj_info, self.variable2, Undef, 10, False),
                       (1, self.obj_info, self.variable1, Undef, 20, False)])
        self.assertEqual(changes2,
                      [(2, self.obj_info, self.variable2, Undef, 10, False),
                       (2, self.obj_info, self.variable1, Undef, 20, False)])

        del changes1[:]
        del changes2[:]

        self.obj.prop1 = None
        self.obj.prop2 = None

        self.assertEqual(changes1,
                      [(1, self.obj_info, self.variable1, 20, None, False),
                       (1, self.obj_info, self.variable2, 10, None, False)])
        self.assertEqual(changes2,
                      [(2, self.obj_info, self.variable1, 20, None, False),
                       (2, self.obj_info, self.variable2, 10, None, False)])

        del changes1[:]
        del changes2[:]

        del self.obj.prop1
        del self.obj.prop2

        self.assertEqual(changes1,
                      [(1, self.obj_info, self.variable1, None, Undef, False),
                       (1, self.obj_info, self.variable2, None, Undef, False)])
        self.assertEqual(changes2,
                      [(2, self.obj_info, self.variable1, None, Undef, False),
                       (2, self.obj_info, self.variable2, None, Undef, False)])

    def test_add_change_notification_with_arg(self):
        changes1 = []
        changes2 = []
        def object_changed1(obj_info, variable,
                            old_value, new_value, fromdb, arg):
            changes1.append((1, obj_info, variable,
                             old_value, new_value, fromdb, arg))
        def object_changed2(obj_info, variable,
                            old_value, new_value, fromdb, arg):
            changes2.append((2, obj_info, variable,
                             old_value, new_value, fromdb, arg))

        self.obj_info.checkpoint()

        obj = object()

        self.obj_info.event.hook("changed", object_changed1, obj)
        self.obj_info.event.hook("changed", object_changed2, obj)

        self.obj.prop2 = 10
        self.obj.prop1 = 20

        self.assertEqual(changes1,
                  [(1, self.obj_info, self.variable2, Undef, 10, False, obj),
                   (1, self.obj_info, self.variable1, Undef, 20, False, obj)])
        self.assertEqual(changes2,
                  [(2, self.obj_info, self.variable2, Undef, 10, False, obj),
                   (2, self.obj_info, self.variable1, Undef, 20, False, obj)])

        del changes1[:]
        del changes2[:]

        self.obj.prop1 = None
        self.obj.prop2 = None

        self.assertEqual(changes1,
                  [(1, self.obj_info, self.variable1, 20, None, False, obj),
                   (1, self.obj_info, self.variable2, 10, None, False, obj)])
        self.assertEqual(changes2,
                  [(2, self.obj_info, self.variable1, 20, None, False, obj),
                   (2, self.obj_info, self.variable2, 10, None, False, obj)])

        del changes1[:]
        del changes2[:]

        del self.obj.prop1
        del self.obj.prop2

        self.assertEqual(changes1,
              [(1, self.obj_info, self.variable1, None, Undef, False, obj),
               (1, self.obj_info, self.variable2, None, Undef, False, obj)])
        self.assertEqual(changes2,
              [(2, self.obj_info, self.variable1, None, Undef, False, obj),
               (2, self.obj_info, self.variable2, None, Undef, False, obj)])

    def test_remove_change_notification(self):
        changes1 = []
        changes2 = []
        def object_changed1(obj_info, variable, old_value, new_value, fromdb):
            changes1.append((1, obj_info, variable,
                             old_value, new_value, fromdb))
        def object_changed2(obj_info, variable, old_value, new_value, fromdb):
            changes2.append((2, obj_info, variable,
                             old_value, new_value, fromdb))

        self.obj_info.checkpoint()

        self.obj_info.event.hook("changed", object_changed1)
        self.obj_info.event.hook("changed", object_changed2)
        self.obj_info.event.unhook("changed", object_changed1)

        self.obj.prop2 = 20
        self.obj.prop1 = 10

        self.assertEqual(changes1, [])
        self.assertEqual(changes2,
                      [(2, self.obj_info, self.variable2, Undef, 20, False),
                       (2, self.obj_info, self.variable1, Undef, 10, False)])

    def test_remove_change_notification_with_arg(self):
        changes1 = []
        changes2 = []
        def object_changed1(obj_info, variable,
                            old_value, new_value, fromdb, arg):
            changes1.append((1, obj_info, variable,
                             old_value, new_value, fromdb, arg))
        def object_changed2(obj_info, variable,
                            old_value, new_value, fromdb, arg):
            changes2.append((2, obj_info, variable,
                             old_value, new_value, fromdb, arg))

        self.obj_info.checkpoint()

        obj = object()

        self.obj_info.event.hook("changed", object_changed1, obj)
        self.obj_info.event.hook("changed", object_changed2, obj)
        self.obj_info.event.unhook("changed", object_changed1, obj)

        self.obj.prop2 = 20
        self.obj.prop1 = 10

        self.assertEqual(changes1, [])
        self.assertEqual(changes2,
                  [(2, self.obj_info, self.variable2, Undef, 20, False, obj),
                   (2, self.obj_info, self.variable1, Undef, 10, False, obj)])

    def test_auto_remove_change_notification(self):
        changes1 = []
        changes2 = []
        def object_changed1(obj_info, variable, old_value, new_value, fromdb):
            changes1.append((1, obj_info, variable,
                             old_value, new_value, fromdb))
            return False
        def object_changed2(obj_info, variable, old_value, new_value, fromdb):
            changes2.append((2, obj_info, variable,
                             old_value, new_value, fromdb))
            return False

        self.obj_info.checkpoint()

        self.obj_info.event.hook("changed", object_changed1)
        self.obj_info.event.hook("changed", object_changed2)

        self.obj.prop2 = 20
        self.obj.prop1 = 10

        self.assertEqual(changes1,
                      [(1, self.obj_info, self.variable2, Undef, 20, False)])
        self.assertEqual(changes2,
                      [(2, self.obj_info, self.variable2, Undef, 20, False)])

    def test_auto_remove_change_notification_with_arg(self):
        changes1 = []
        changes2 = []
        def object_changed1(obj_info, variable,
                            old_value, new_value, fromdb, arg):
            changes1.append((1, obj_info, variable,
                             old_value, new_value, fromdb, arg))
            return False
        def object_changed2(obj_info, variable,
                            old_value, new_value, fromdb, arg):
            changes2.append((2, obj_info, variable,
                             old_value, new_value, fromdb, arg))
            return False

        self.obj_info.checkpoint()

        obj = object()

        self.obj_info.event.hook("changed", object_changed1, obj)
        self.obj_info.event.hook("changed", object_changed2, obj)

        self.obj.prop2 = 20
        self.obj.prop1 = 10

        self.assertEqual(changes1,
                  [(1, self.obj_info, self.variable2, Undef, 20, False, obj)])
        self.assertEqual(changes2,
                  [(2, self.obj_info, self.variable2, Undef, 20, False, obj)])

    def test_get_obj(self):
        self.assertTrue(self.obj_info.get_obj() is self.obj)

    def test_get_obj_reference(self):
        """
        We used to assign the get_obj() manually. This breaks stored
        references to the method (IOW, what we do in the test below).

        It was a bit faster, but in exchange for the danger of introducing
        subtle bugs which are super hard to debug.
        """
        get_obj = self.obj_info.get_obj
        self.assertTrue(get_obj() is self.obj)
        another_obj = self.Class()
        self.obj_info.set_obj(another_obj)
        self.assertTrue(get_obj() is another_obj)

    def test_set_obj(self):
        obj = self.Class()
        self.obj_info.set_obj(obj)
        self.assertTrue(self.obj_info.get_obj() is obj)

    def test_weak_reference(self):
        obj = self.Class()
        obj_info = get_obj_info(obj)
        del obj
        self.assertEqual(obj_info.get_obj(), None)

    def test_object_deleted_notification(self):
        obj = self.Class()
        obj_info = get_obj_info(obj)
        obj_info["tainted"] = True
        deleted = []
        def object_deleted(obj_info):
            deleted.append(obj_info)
        obj_info.event.hook("object-deleted", object_deleted)
        del obj_info
        del obj
        self.assertEqual(len(deleted), 1)
        self.assertTrue("tainted" in deleted[0])

    def test_object_deleted_notification_after_set_obj(self):
        obj = self.Class()
        obj_info = get_obj_info(obj)
        obj_info["tainted"] = True
        obj = self.Class()
        obj_info.set_obj(obj)
        deleted = []
        def object_deleted(obj_info):
            deleted.append(obj_info)
        obj_info.event.hook("object-deleted", object_deleted)
        del obj_info
        del obj
        self.assertEqual(len(deleted), 1)
        self.assertTrue("tainted" in deleted[0])


class ClassAliasTest(TestHelper):

    def setUp(self):
        TestHelper.setUp(self)
        class Class:
            __storm_table__ = "table"
            prop1 = Property("column1", primary=True)
        self.Class = Class
        self.ClassAlias = ClassAlias(self.Class, "alias")

    def test_cls_info_cls(self):
        cls_info = get_cls_info(self.ClassAlias)
        self.assertEqual(cls_info.cls, self.Class)
        self.assertEqual(cls_info.table.name, "alias")
        self.assertEqual(self.ClassAlias.prop1.name, "column1")
        self.assertEqual(self.ClassAlias.prop1.table, self.ClassAlias)

    def test_compile(self):
        statement = compile(self.ClassAlias)
        self.assertEqual(statement, "alias")

    def test_compile_with_reserved_keyword(self):
        Alias = ClassAlias(self.Class, "select")
        statement = compile(Alias)
        self.assertEqual(statement, '"select"')

    def test_compile_in_select(self):
        expr = Select(self.ClassAlias.prop1, self.ClassAlias.prop1 == 1,
                      self.ClassAlias)
        statement = compile(expr)
        self.assertEqual(statement,
                         'SELECT alias.column1 FROM "table" AS alias '
                         'WHERE alias.column1 = ?')

    def test_compile_in_select_with_reserved_keyword(self):
        Alias = ClassAlias(self.Class, "select")
        expr = Select(Alias.prop1, Alias.prop1 == 1, Alias)
        statement = compile(expr)
        self.assertEqual(statement,
                         'SELECT "select".column1 FROM "table" AS "select" '
                         'WHERE "select".column1 = ?')

    def test_crazy_metaclass(self):
        """We don't want metaclasses playing around when we build an alias."""
        TestHelper.setUp(self)
        class MetaClass(type):
            def __new__(meta_cls, name, bases, dict):
                cls = type.__new__(meta_cls, name, bases, dict)
                cls.__storm_table__ = "HAH! GOTCH YA!"
                return cls
        class Class(metaclass=MetaClass):
            __storm_table__ = "table"
            prop1 = Property("column1", primary=True)
        Alias = ClassAlias(Class, "USE_THIS")
        self.assertEqual(Alias.__storm_table__, "USE_THIS")

    def test_cached_aliases(self):
        """
        Class aliases are cached such that multiple invocations of
        C{ClassAlias} return the same object.
        """
        alias1 = ClassAlias(self.Class, "something_unlikely")
        alias2 = ClassAlias(self.Class, "something_unlikely")
        self.assertIdentical(alias1, alias2)
        alias3 = ClassAlias(self.Class, "something_unlikely2")
        self.assertNotIdentical(alias1, alias3)
        alias4 = ClassAlias(self.Class, "something_unlikely2")
        self.assertIdentical(alias3, alias4)

    def test_unnamed_aliases_not_cached(self):
        alias1 = ClassAlias(self.Class)
        alias2 = ClassAlias(self.Class)
        self.assertNotIdentical(alias1, alias2)

    def test_alias_cache_is_per_class(self):
        """
        The cache of class aliases is not as bad as it once was.
        """
        class LocalClass(self.Class):
            pass
        alias = ClassAlias(self.Class, "something_unlikely")
        alias2 = ClassAlias(LocalClass, "something_unlikely")
        self.assertNotIdentical(alias, alias2)

    def test_aliases_only_last_as_long_as_class(self):
        """
        The cached ClassAliases only last for as long as the class is alive.
        """
        class LocalClass(self.Class):
            pass
        alias = ClassAlias(LocalClass, "something_unlikely3")
        alias_ref = ref(alias)
        class_ref = ref(LocalClass)
        del alias
        del LocalClass

        gc.collect(); gc.collect(); gc.collect()

        self.assertIdentical(class_ref(), None)
        self.assertIdentical(alias_ref(), None)


class TypeCompilerTest(TestHelper):

    def test_nested_classes(self):
        """Convoluted case checking that the model is right."""
        class Class1:
            __storm_table__ = "class1"
            id = Property(primary=True)
        class Class2:
            __storm_table__ = Class1
            id = Property(primary=True)
        statement = compile(Class2)
        self.assertEqual(statement, "class1")
        alias = ClassAlias(Class2, "alias")
        statement = compile(Select(alias.id))
        self.assertEqual(statement, "SELECT alias.id FROM class1 AS alias")




from unittest import defaultTestLoader

from storm.properties import Int
from storm.info import get_obj_info
from storm.cache import Cache, GenerationalCache
from storm.tests.helper import TestHelper


class StubObjectInfo:

    def __init__(self, id):
        self.id = id
        self.hashed = False

    def get_obj(self):
        return str(self.id)

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self.id)

    def __hash__(self):
        self.hashed = True
        return self.id

    def __lt__(self, other):
        return self.id < other.id


class StubClass:

    __storm_table__ = "stub_class"

    id = Int(primary=True)


class BaseCacheTest(TestHelper):

    Cache = Cache

    def setUp(self):
        super().setUp()
        self.obj_infos = [StubObjectInfo(i) for i in range(10)]
        for i in range(len(self.obj_infos)):
            setattr(self, "obj%d" % (i+1), self.obj_infos[i])

    def clear_hashed(self):
        for obj_info in self.obj_infos:
            obj_info.hashed = False

    def test_initially_empty(self):
        cache = self.Cache()
        self.assertEqual(cache.get_cached(), [])

    def test_add(self):
        cache = self.Cache(5)
        cache.add(self.obj1)
        cache.add(self.obj2)
        cache.add(self.obj3)
        self.assertEqual(sorted(cache.get_cached()),
                         [self.obj1, self.obj2, self.obj3])

    def test_adding_similar_obj_infos(self):
        """If __eq__ is broken, this fails."""
        obj_info1 = get_obj_info(StubClass())
        obj_info2 = get_obj_info(StubClass())
        cache = self.Cache(5)
        cache.add(obj_info1)
        cache.add(obj_info2)
        cache.add(obj_info2)
        cache.add(obj_info1)
        self.assertEqual(sorted([hash(obj_info)
                                 for obj_info in cache.get_cached()]),
                         sorted([hash(obj_info1), hash(obj_info2)]))

    def test_remove(self):
        cache = self.Cache(5)
        cache.add(self.obj1)
        cache.add(self.obj2)
        cache.add(self.obj3)
        cache.remove(self.obj2)
        self.assertEqual(sorted(cache.get_cached()),
                         [self.obj1, self.obj3])

    def test_add_existing(self):
        cache = self.Cache(5)
        cache.add(self.obj1)
        cache.add(self.obj2)
        cache.add(self.obj3)
        cache.add(self.obj2)
        self.assertEqual(sorted(cache.get_cached()),
                         [self.obj1, self.obj2, self.obj3])

    def test_add_with_size_zero(self):
        """Cache is disabled entirely on add() if size is 0."""
        cache = self.Cache(0)
        cache.add(self.obj1)
        # Ensure that we don't even check if obj_info is in the
        # cache, by testing if it was hashed.  Hopefully, that means
        # we got a faster path.
        self.assertEqual(self.obj1.hashed, False)

    def test_remove_with_size_zero(self):
        """Cache is disabled entirely on remove() if size is 0."""
        cache = self.Cache(0)
        cache.remove(self.obj1)

    def test_clear(self):
        """The clear method empties the cache."""
        cache = self.Cache(5)
        for obj_info in self.obj_infos:
            cache.add(obj_info)
        cache.clear()
        self.assertEqual(cache.get_cached(), [])

        # Just an additional check ensuring that any additional structures
        # which may be used were cleaned properly as well.
        for obj_info in self.obj_infos:
            self.assertEqual(cache.remove(obj_info), False)

    def test_set_zero_size(self):
        """
        Setting a cache's size to zero clears the cache.
        """
        cache = self.Cache()
        cache.add(self.obj1)
        cache.add(self.obj2)
        cache.set_size(0)
        self.assertEqual(cache.get_cached(), [])

    def test_fit_size(self):
        """
        A cache of size n can hold at least n objects.
        """
        size = 10
        cache = self.Cache(size)
        for value in range(size):
            cache.add(StubObjectInfo(value))
        self.assertEqual(len(cache.get_cached()), size)


class CacheTest(BaseCacheTest):

    def test_size_and_fifo_behaviour(self):
        cache = Cache(5)
        for obj_info in self.obj_infos:
            cache.add(obj_info)
        self.assertEqual([obj_info.id for obj_info in cache.get_cached()],
                         [9, 8, 7, 6, 5])

    def test_reduce_max_size_to_zero(self):
        """When setting the size to zero, there's an optimization."""
        cache = Cache(5)
        obj_info = self.obj_infos[0]
        cache.add(obj_info)
        obj_info.hashed = False
        cache.set_size(0)
        self.assertEqual(cache.get_cached(), [])
        # Ensure that we don't even check if obj_info is in the
        # cache, by testing if it was hashed.  Hopefully, that means
        # we got a faster path.
        self.assertEqual(obj_info.hashed, False)

    def test_reduce_max_size(self):
        cache = Cache(5)
        for obj_info in self.obj_infos:
            cache.add(obj_info)
        cache.set_size(3)
        self.assertEqual([obj_info.id for obj_info in cache.get_cached()],
                         [9, 8, 7])

        # Adding items past the new maximum size should drop older ones.
        for obj_info in self.obj_infos[:2]:
            cache.add(obj_info)
        self.assertEqual([obj_info.id for obj_info in cache.get_cached()],
                         [1, 0, 9])

    def test_increase_max_size(self):
        cache = Cache(5)
        for obj_info in self.obj_infos:
            cache.add(obj_info)
        cache.set_size(10)
        self.assertEqual([obj_info.id for obj_info in cache.get_cached()],
                         [9, 8, 7, 6, 5])

        # Adding items past the new maximum size should drop older ones.
        for obj_info in self.obj_infos[:6]:
            cache.add(obj_info)
        self.assertEqual([obj_info.id for obj_info in cache.get_cached()],
                         [5, 4, 3, 2, 1, 0, 9, 8, 7, 6])


class TestGenerationalCache(BaseCacheTest):

    Cache = GenerationalCache

    def setUp(self):
        super().setUp()
        self.obj1 = StubObjectInfo(1)
        self.obj2 = StubObjectInfo(2)
        self.obj3 = StubObjectInfo(3)
        self.obj4 = StubObjectInfo(4)

    def test_initially_empty(self):
        cache = GenerationalCache()
        self.assertEqual(cache.get_cached(), [])

    def test_cache_one_object(self):
        cache = GenerationalCache()
        cache.add(self.obj1)
        self.assertEqual(cache.get_cached(), [self.obj1])

    def test_cache_multiple_objects(self):
        cache = GenerationalCache()
        cache.add(self.obj1)
        cache.add(self.obj2)
        self.assertEqual(sorted(cache.get_cached()), [self.obj1, self.obj2])

    def test_clear_cache(self):
        cache = GenerationalCache()
        cache.add(self.obj1)
        cache.clear()
        self.assertEqual(cache.get_cached(), [])

    def test_clear_cache_clears_the_second_generation(self):
        cache = GenerationalCache(1)
        cache.add(self.obj1)
        cache.add(self.obj2)
        cache.clear()
        self.assertEqual(cache.get_cached(), [])

    def test_remove_object(self):
        cache = GenerationalCache()
        cache.add(self.obj1)
        cache.add(self.obj2)
        cache.add(self.obj3)

        present = cache.remove(self.obj2)
        self.assertTrue(present)
        self.assertEqual(sorted(cache.get_cached()), [self.obj1, self.obj3])

    def test_remove_nothing(self):
        cache = GenerationalCache()
        cache.add(self.obj1)

        present = cache.remove(self.obj2)
        self.assertFalse(present)
        self.assertEqual(cache.get_cached(), [self.obj1])

    def test_size_limit(self):
        """
        A cache will never hold more than twice its size in objects.  The
        generational system is what prevents it from holding exactly the
        requested number of objects.
        """
        size = 10
        cache = GenerationalCache(size)
        for value in range(5 * size):
            cache.add(StubObjectInfo(value))
        self.assertEqual(len(cache.get_cached()), size * 2)

    def test_set_size_smaller_than_current_size(self):
        """
        Setting the size to a smaller size than the number of objects
        currently cached will drop some of the extra content.  Note that
        because of the generation system, it can actually hold two times
        the size requested in edge cases.
        """
        cache = GenerationalCache(150)
        for i in range(250):
            cache.add(StubObjectInfo(i))
        cache.set_size(100)
        cached = cache.get_cached()
        self.assertEqual(len(cached), 100)
        for obj_info in cache.get_cached():
            self.assertTrue(obj_info.id >= 100)

    def test_set_size_larger_than_current_size(self):
        """
        Setting the cache size to something more than the number of
        objects in the cache does not affect its current contents,
        and will merge any elements from the second generation into
        the first one.
        """
        cache = GenerationalCache(1)
        cache.add(self.obj1) # new=[1]    old=[]
        cache.add(self.obj2) # new=[2]    old=[1]
        cache.set_size(2)    # new=[1, 2] old=[]
        cache.add(self.obj3) # new=[3]    old=[1, 2]
        self.assertEqual(sorted(cache.get_cached()),
                         [self.obj1, self.obj2, self.obj3])

    def test_set_size_limit(self):
        """
        Setting the size limits the cache's size just like passing an
        initial size would.
        """
        size = 10
        cache = GenerationalCache(size * 100)
        cache.set_size(size)
        for value in range(size * 10):
            cache.add(StubObjectInfo(value))
        self.assertEqual(len(cache.get_cached()), size * 2)

    def test_two_generations(self):
        """
        Inserting more objects than the cache's size causes the cache
        to contain two generations, each holding up to <size> objects.
        """
        cache = GenerationalCache(1)
        cache.add(self.obj1)
        cache.add(self.obj2)

        self.assertEqual(sorted(cache.get_cached()), [self.obj1, self.obj2])

    def test_three_generations(self):
        """
        If more than 2*<size> objects come along, only 2*<size>
        objects are retained.
        """
        cache = GenerationalCache(1)
        cache.add(self.obj1)
        cache.add(self.obj2)
        cache.add(self.obj3)

        self.assertEqual(sorted(cache.get_cached()), [self.obj2, self.obj3])

    def test_generational_overlap(self):
        """
        An object that is both in the primary and the secondary
        generation is listed only once in the cache's contents.
        """
        cache = GenerationalCache(2)
        cache.add(self.obj1) # new=[1]    old=[]
        cache.add(self.obj2) # new=[1, 2] old=[]
        cache.add(self.obj3) # new=[3]    old=[1, 2]
        cache.add(self.obj1) # new=[3, 1] old=[1, 2]

        self.assertEqual(sorted(cache.get_cached()),
                         [self.obj1, self.obj2, self.obj3])

    def test_remove_from_overlap(self):
        """
        Removing an object from the cache removes it from both its
        primary and secondary generations.
        """
        cache = GenerationalCache(2)
        cache.add(self.obj1) # new=[1]    old=[]
        cache.add(self.obj2) # new=[1, 2] old=[]
        cache.add(self.obj3) # new=[3]    old=[1, 2]
        cache.add(self.obj1) # new=[3, 1] old=[1, 2]

        present = cache.remove(self.obj1)
        self.assertTrue(present)
        self.assertEqual(sorted(cache.get_cached()), [self.obj2, self.obj3])

    def test_evict_oldest(self):
        """The "oldest" object is the first to be evicted."""
        cache = GenerationalCache(1)
        cache.add(self.obj1)
        cache.add(self.obj2)
        cache.add(self.obj3)

        self.assertEqual(sorted(cache.get_cached()), [self.obj2, self.obj3])

    def test_evict_LRU(self):
        """
        Actually, it's not the oldest but the LRU object that is first
        to be evicted.  Re-adding the oldest object makes it not be
        the LRU.
        """
        cache = GenerationalCache(1)
        cache.add(self.obj1)
        cache.add(self.obj2)

        # This "refreshes" the oldest object in the cache.
        cache.add(self.obj1)

        cache.add(self.obj3)

        self.assertEqual(sorted(cache.get_cached()), [self.obj1, self.obj3])


def test_suite():
    return defaultTestLoader.loadTestsFromName(__name__)



#
# Copyright (c) 2006, 2007 Canonical
#
# Written by Gustavo Niemeyer <gustavo@niemeyer.net>
#
# This file is part of Storm Object Relational Mapper.
#
# Storm is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation; either version 2.1 of
# the License, or (at your option) any later version.
#
# Storm is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from zope.interface import classImplements

from storm.info import ObjectInfo
from storm.store import EmptyResultSet, ResultSet
from storm.zope.interfaces import IResultSet, ISQLObjectResultSet
from storm import sqlobject as storm_sqlobject


classImplements(storm_sqlobject.SQLObjectResultSet, ISQLObjectResultSet)
classImplements(ResultSet, IResultSet)
classImplements(EmptyResultSet, IResultSet)

try:
    from zope.security.checker import NoProxy, BasicTypes, _available_by_default
except ImportError:
    # We don't have zope.security installed.
    pass
else:
    # The following is required for storm.info.get_obj_info() to have
    # access to a proxied object which is already in the store (IOW, has
    # the object info set already).  With this, Storm is able to
    # gracefully handle situations when a proxied object is passed to a
    # Store.
    _available_by_default.append("__storm_object_info__")
    BasicTypes[ObjectInfo] = NoProxy
